"""The XMLDataHandler handles all XML data needed for making
recommendations."""
from data_handler import DataHandler
from urllib import parse
from urllib import request


class XMLDataHandler(DataHandler):
    """ Handle all XML formatted data needed for recommendations"""
    def __init__(self, source):
        self._source = source

    def get_content(self):
        raise NotImplementedError

    def get_top_matching_items(self, item, limit):
        raise NotImplementedError

    def update_top_matching_items(self, top_matches):
        raise NotImplementedError

    def _fetch_data(self, path):
        """Fetches XML encoded data and returns it as raw XML.

         parameters:
             path The path at which the resource can be reached"""
        data = request.Request(self._source + path)
        return request.urlopen(data).read().decode("utf-8")

    def _post_data(self, data, path, key="data"):
        """Posts XML encoded data to a given path.

         parameters:
             data The XML data to post
             path The path at which the data should be posted
             key The POST parameter name to use. Default is data"""
        data = parse.urlencode({key: data})
        update_request = request.Request(self._source + path)
        update_request.add_header("Content-Type",
                                  "application/xml;charset=utf-8")
        request.urlopen(update_request, data.encode("utf-8"))




"""Contains similarity algorithms used to determine how similar two items are.
Euclidean distance and Pearson corellation"""


def euclidean_distance(item1, item2):
    """Calculates a similarity rating between two data items. Returns
    a float value between 0 and 1. A value of 1 represents a high
    similarity rating, a value of 0 represents the lowest
    similarity rating.

    arguments:
    item1 & item2 - the itmes being compared
    """
    common_properties = __get_common_properties(item1, item2)
    sum_of_squares = 0
    if common_properties:
        for property in common_properties:
            sum_of_squares += (item2[property] - item1[property])**2
        return 1 / (1 + sum_of_squares**0.5)
    else:
        return 0


def pearson_corellation(item1, item2):
    """Calculate the pearson corellation coefficient between two items.
    Return a float representing the corellation coefficient. More info can
    be found here:
    https://en.wikipedia.org/wiki/Pearson_product-moment_correlation_coefficient
    """
    common_properties = __get_common_properties(item1, item2)
    item_count = len(common_properties)
    if item_count > 0:
        item1 = {k: item1[k] for k in common_properties}
        item2 = {k: item2[k] for k in common_properties}

        item1_sum = sum(item1.values())
        item2_sum = sum(item2.values())
    if item_count > 0:
        numerator = 0

        item1_squared_sum = 0
        item2_squared_sum = 0
        for key in item1.keys():
            factor_item1 = __factor(item1[key], item1_sum, item_count)
            factor_item2 = __factor(item2[key], item2_sum, item_count)

            numerator += factor_item1 * factor_item2
            item1_squared_sum += factor_item1**2
            item2_squared_sum += factor_item2**2

        return numerator / (item1_squared_sum**0.5 * item2_squared_sum**0.5)

    else:
        return 0

def tanimoto_coefficient(item1, item2):
    """Applicable in cases when similarity is calculated based on the
    presense or absense of a given characteristic. It's required that
    item1 and item2 have bitmaps of the characteristics. E.g. 1 corresponds
    to the presense of a characteristic - 0 to the absence."""

    common_properties = __get_common_properties(item1, item2)
    item_count = len(common_properties)
    if item_count == 0:
        return 0

    return float(item_count/(len(item1) + len(item2) - item_count))


def __factor(score, items_sum, item_count):
    """Helper method for the pearson correlation coefficient algorithm."""
    return score - items_sum/item_count


def __get_common_properties(item1, item2):
    """Return the common keys that two maps have"""
    return set(item1.keys()) & set(item2.keys())



"""Main Recommender module."""


class Recommender(object):
    """Performs recommendations based on supplied data and a comparison
    algorithm.

    Attributes:
        data_provider         DataProvider instance
        similarity_algorithm  SimilarityMetric instance"""
    def __init__(self, data_provider, similarity_metric):
        self.__provider = data_provider
        self.__similarity = similarity_metric

    def item_based_recommendations(self, subject):
        """Build a list of recommendations relevant to the items's
        properties. Returns a list of tuples where the first item is
        the anticipated score the subject might give the recommended item
        the second is the recommended item.

        arguments:
            item The item for which recommendations are being built"""
        data = self.__provider.get_content()
        total_scores = {}
        score_sums = {}
        top_similarities = self.__provider.get_top_matching_items(subject)

        if top_similarities == {}:
            top_similarities = self.calculate_similarities(data, subject, 10)

        for similarity, other in top_similarities:
            if similarity <= 0:
                continue

            for item in data[other].keys():
                if item not in data[subject] or data[subject][item] == 0:
                    total_scores.setdefault(item, 0)
                    score_sums.setdefault(item, 0)

                    total_scores[item] += data[other][item] * similarity
                    score_sums[item] += similarity

        rankings = [(total / score_sums[item], item)
                    for item, total in total_scores.items()]
        return sorted(rankings, key=lambda x: x[0], reverse=True)

    def content_based_recommendations(self, subject):
        """Build a list of recommendations for a given subject based on
        the most similar content to the subject's properties. Returns a
        list of tuples where the first item is the anticipated score the
        subject might give the recommended item the second is the
        recommended item.

        arguments:
            item The item for which recommendations are being built"""
        data = self.__provider.get_content()
        transposed = self.__provider.transposed_content()

        content_similarities = self.__provider.get_top_matching_items()

        if content_similarities == {}:
            content_similarities = self.calculate_similarities(transposed,
                                                               limit=10)
        total_scores = {}
        score_sums = {}

        for item, rating in data[subject].items():
            for similarity, item2 in content_similarities[item]:
                if item2 in data[subject]:
                    continue

                total_scores.setdefault(item2, 0)
                score_sums.setdefault(item2, 0)

                total_scores[item2] += similarity * rating
                score_sums[item2] += similarity

        rankings = [(score / score_sums[item], item)
                    for item, score in total_scores.items()]
        return sorted(rankings, key=lambda x: x[0], reverse=True)

    def calculate_similarities(self, data, subject="", limit=3):
        """Generate a dictionary in which each key is an item and the
        value is a list of its top matches, e.g.:
        {"<item>": [(<similarity_score>, <similar_item>)]}

        arguments:
            data The poll from which to calculate the most similar items
            subject The subject we want to return results for. If left blank all
                    results are returned.
            limit The max number of top matching items returnedi. If 0 all
                  reslts are returned."""
        result = {}
        for item in data.keys():
            result[item] = self.__top_matches(item, data, limit)
        if subject == "":
            return result
        return result[subject]

    def __top_matches(self, item, data, limit=3):
        """Generate a list of top matching items to another item. The list
        consists of tuples: (<similarity_score>, <item>)

        arguments:
            item The item for which the top matches are calculated
            data The dataset from which scores are pulled
            limit The max number of top matching items returned, if 0 all
                  results are returned"""
        scores = [(self.__similarity(data[item], data[other]), other)
                  for other in data.keys() if other != item]
        if limit == 0:
            return sorted(scores, key=lambda x: x[0], reverse=True)
        return sorted(scores, key=lambda x: x[0], reverse=True)[:limit]



"""The JSONDataHandler handles all JSON data needed for making
recommendations."""
from data_handler import DataHandler
from urllib import parse
from urllib import request
import json


class JSONDataHandler(DataHandler):
    """Handle all JSON data needed for recommendations"""

    def __init__(self, source):
        self._source = source

    def get_content(self):
        raise NotImplementedError

    def get_top_matching_items(self, item, limit):
        raise NotImplementedError

    def update_top_matching_items(self, top_matches):
        raise NotImplementedError

    def _fetch_data(self, path):
        """Fetches JSON encoded data and returns it as a dictionary.

        parameters:
            path The at which the resource can be reached"""
        data = request.Request(self._source + path)
        return json.loads(request.urlopen(data).read().decode("utf-8"))

    def _post_data(self, data, path, key="data"):
        """Posts JSON encoded data to a given path.

        parameters:
            data The data to post
            path The path at which the data should be posted
            key The POST parameter name to use. Default is data"""
        data = parse.urlencode({key: json.dumps(data)})

        update_request = request.Request(self._source + path)
        update_request.add_header("Content-Type",
                                  "application/json;charset=utf-8")
        request.urlopen(update_request, data.encode("utf-8"))



from setuptools import setup, find_packages
# To use a consistent encoding
from codecs import open
from os import path

here = path.abspath(path.dirname(__file__))

# Get the long description from the relevant file
with open(path.join(here, 'DESCRIPTION.rst'), encoding='utf-8') as f:
    long_description = f.read()

setup(
    name='recommendy',

    version='0.0.1',

    description='Python3 recommendation system',
    long_description=long_description,

    url='https://github.com/d0ivanov/recommendy',

    author='Dobromir Ivanov',
    author_email='dobromir0ivanov@gmail.com',

    license='MIT',

    #https://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=[
        'Intended Audience :: Developers',
        'Topic :: Internet :: WWW/HTTP :: Indexing/Search',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
    ],

    # What does your project relate to?
    keywords='recommendation',

    # You can just specify the packages manually here if your project is
    # simple. Or you can use find_packages().
    packages=find_packages(exclude=['contrib', 'docs', 'tests*']),

    # List run-time dependencies here.  These will be installed by pip when
    # your project is installed. For an analysis of "install_requires" vs pip's
    # requirements files see:
    # https://packaging.python.org/en/latest/requirements.html
    install_requires=['json', 'lxml'],

    # List additional groups of dependencies here (e.g. development
    # dependencies). You can install these using the following syntax,
    # for example:
    # $ pip install -e .[dev,test]
    extras_require={
        'dev': ['check-manifest'],
        'test': ['coverage'],
    },

    # If there are data files included in your packages that need to be
    # installed, specify them here.  If using Python 2.6 or less, then these
    # have to be included in MANIFEST.in as well.
    package_data={},

    # Although 'package_data' is the preferred approach, in some case you may
    # need to place data files outside of your packages. See:
    # http://docs.python.org/3.4/distutils/setupscript.html#installing-additional-files # noqa
    # In this case, 'data_file' will be installed into '<sys.prefix>/my_data'
    data_files=[],

    # To provide executable scripts, use entry points in preference to the
    # "scripts" keyword. Entry points provide cross-platform support and allow
    # pip to create the appropriate form of executable for the target platform.
    entry_points={
        'console_scripts':[]
    }
)



"""Contains data handling functionality. The DataHandler class is resposible
for serving all data necessary for the recommender"""


class DataHandler(object):
    """Represents the interface each data provider class should implement."""

    def get_content(self):
        """Returns a dictionary containing the data on which recommendations
        are based."""
        raise NotImplementedError

    def get_top_matching_items(self, item, limit):
        """Returns a dictionary containing the top matching items to
        a given item.

        parameters:
            item The item for which the top matches are returned
            limit The max number of results returned, if 0 should
                  return all results found"""
        raise NotImplementedError

    def update_top_matching_items(self, top_matches):
        """Save updated data for item similarities

        parameters:
            top_mathces A dictionary containing items that should be updated"""
        raise NotImplementedError

    def transposed_content(self):
        """Transpose the content so that properties become keys
        and keys become properties."""
        result = {}
        for key, value in self.get_content().items():
            for property, score in value.items():
                result.setdefault(property, {})
                result[property][key] = score
        return result
